# -*- coding: utf8 -*-

from collections import defaultdict, namedtuple
from typing import List, Tuple, Dict, Union, FrozenSet, Iterable

from .influence_graph import InfluenceGraph, Gene, Process



class Transition(namedtuple('Transition', 'gene process states')):
    def __str__(self):
        return f'{self.gene.name} {set(p.name for p in self.process) or "{}"} → {set(self.states) or "{}"}'


class DiscreteModel:
    def __init__(self, influence_graph: InfluenceGraph):
        self.influence_graph: InfluenceGraph = influence_graph
        self.transitions: List[Transition] = []

    def add_transition(self, gene: str, process: Tuple[str], states: Tuple[int]):
        gene = self.influence_graph.find_gene_by_name(gene)
        process = tuple(self.influence_graph.find_process_by_name(p) for p in process)
        self.transitions.append(Transition(gene, process, states))
    
    def find_transition(self, gene: Gene, process: Tuple[Process]):
        for transition in self.transitions:
            if transition.gene == gene and set(transition.process) == set(process):
                return transition
        raise AttributeError(f'transition "{gene.name} {set(p.name for p in process) or "{}"}" does not exist')

    def list_transitions(self) -> Tuple[Transition]:
        return tuple(self.transitions)

    def available_state(self, gene: Gene, states: Dict[Gene, int]) -> Tuple[int]:
        return tuple(self._available_state(gene, states))
    
    def _available_state(self, gene: Gene, states: Dict[Gene, int]) -> Iterable[int]:
        done = set()
        active_process = gene.active_process(states)
        transition = self.find_transition(gene, active_process)
        current_state = states[gene]
        for state in transition.states:
            next_state = current_state + (current_state < state) - (current_state > state)
            if next_state not in done:
                done.add(next_state)
                yield next_state

    def __str__(self):
        string = str(self.influence_graph)
        string += '\n\tmodel'
        for transition in self.list_transitions():
            string += f'\n\t\t{transition}'
        return string

    __repr__ = __str__

    def __eq__(self, other):
        if not isinstance(other, DiscreteModel):
            return False
        return (self.influence_graph == other.influence_graph
                and self.transitions == other.transitions)



import unittest
from discrete_model import parse_smbionet_output_file, InfluenceGraph, DiscreteModel


class Test(unittest.TestCase):
    def test_process_activation(self):
        model1, _ = parse_smbionet_output_file('resources/mucusOperonV3.out')
        free = model1.influence_graph.find_process_by_name('free')
        mucuB = model1.influence_graph.find_gene_by_name('mucuB')
        self.assertTrue(free.is_active({mucuB: 0}))

    def test_find_transition(self):
        model1, model2 = parse_smbionet_output_file('resources/mucusOperonV3.out')
        operon = model1.influence_graph.find_gene_by_name('operon')
        free = model1.influence_graph.find_process_by_name('free')
        alg = model1.influence_graph.find_process_by_name('alg')
        transition = model1.find_transition(operon, (alg, free))
        self.assertEqual((2,), transition.states)
    
        transition = model2.find_transition(operon, (free,))
        self.assertEqual((1, 2), transition.states)
    
    def test_available_states(self):
        _, model2 = parse_smbionet_output_file('resources/mucusOperonV3.out')
        operon = model2.influence_graph.find_gene_by_name('operon')
        mucuB = model2.influence_graph.find_gene_by_name('mucuB')
        free = model2.influence_graph.find_process_by_name('free')
        self.assertEqual((1,), model2.available_state(operon, {operon: 0, mucuB: 0}))
        self.assertEqual((2,), model2.available_state(operon, {operon: 1, mucuB: 0}))
        self.assertEqual((2,), model2.available_state(operon, {operon: 2, mucuB: 0}))

    def test_mucus_operon_v3(self):
        model1, model2 = parse_smbionet_output_file('resources/mucusOperonV3.out')
        graph = InfluenceGraph()
        graph.add_gene('operon', 0, 2)
        graph.add_gene('mucuB', 0, 1)
        graph.add_process('free', '(not(mucuB>=1))', 'operon')
        graph.add_process('alg', '(operon>=1)', 'operon')
        graph.add_process('prod', '(operon>=1)', 'mucuB')

        expected_model1 = DiscreteModel(graph)
        expected_model1.add_transition('operon', (), (0,))
        expected_model1.add_transition('operon', ('alg',), (2,))
        expected_model1.add_transition('operon', ('free',), (0,))
        expected_model1.add_transition('operon', ('alg', 'free'), (2,))
        expected_model1.add_transition('mucuB', (), (0,))
        expected_model1.add_transition('mucuB', ('prod',), (1,))

        expected_model2 = DiscreteModel(graph)
        expected_model2.add_transition('operon', (), (0,))
        expected_model2.add_transition('operon', ('alg',), (2,))
        expected_model2.add_transition('operon', ('free',), (1, 2))
        expected_model2.add_transition('operon' , ('alg', 'free'), (2,))
        expected_model2.add_transition('mucuB', (), (0,))
        expected_model2.add_transition('mucuB', ('prod',), (1,))

        self.assertEqual(expected_model1, model1)
        self.assertEqual(expected_model2, model2)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf8 -*-

from typing import List, Dict, Tuple, Set
from collections import defaultdict, namedtuple
from dataclasses import dataclass


class Gene:
    def __init__(self, name, states):
        self.name = name
        self.states = states
        self.process = []

    def add_process(self, process: 'Process'):
        if process not in self.process:
            self.process.append(process)
    
    def active_process(self, state: Dict['Gene', int]) -> Tuple['Process']:
        return tuple(process for process in self.process if process.is_active(state))

    def __eq__(self, other):
        if not isinstance(other, Gene):
            return False
        return self.name == other.name and self.states == other.states
    
    def __hash__(self):
        return hash((self.name, self.states))

    def __str__(self):
        return f'{self.name} = {" ".join(map(str, self.states))}'
    
    def __repr__(self):
        return self.name


class Expression(namedtuple('Expression', 'expression')):
    def evaluate(self, **params: int):
        return eval(self.expression, params)
    
    def __str__(self):
        return self.expression


class Process(namedtuple('Process', 'name genes expression')):
    def __init__(self, name: str, genes: Tuple[Gene], expression: Expression):
        for gene in genes:
            gene.add_process(self)

    def is_active(self, states: Dict[Gene, int]):
        params = {gene.name: state for gene, state in states.items()}
        return self.expression.evaluate(**params)

    def __str__(self):
        return f'{self.name} : {self.expression} → {" ".join([gene.name for gene in self.genes])}'


class InfluenceGraph:
    def __init__(self):
        self.genes: List[Gene] = []
        self.process: List[Process] = []

    def add_gene(self, name: str, state_min: int, state_max: int):
        self.genes.append(Gene(name, tuple(range(state_min, state_max+1))))

    def find_gene_by_name(self, gene_name: str) -> Gene:
        for gene in self.genes:
            if gene.name == gene_name:
                return gene
        raise AttributeError(f'gene "{gene_name}" does not exist')

    def list_genes(self) -> Tuple[Gene]:
        return tuple(self.genes)

    def add_process(self, name: str, expression: str, *genes: str):
        genes = tuple(self.find_gene_by_name(gene) for gene in genes)
        process = Process(name, genes, Expression(expression))
        self.process.append(process)

    def find_process_by_name(self, process_name: str) -> Process:
        for process in self.process:
            if process.name == process_name:
                return process
        raise AttributeError(f'process "{process_name}" does not exist')

    def list_process(self) -> Tuple[Process]:
        return tuple(self.process)

    def __str__(self):
        string = '\nInfluenceGraph\n\tgenes:\n\t\t'
        string += '\n\t\t'.join(map(str, self.list_genes()))
        string += '\n\tprocess\n\t\t'
        string += '\n\t\t'.join(map(str, self.list_process()))
        return string

    __repr__ = __str__

    def __eq__(self, other):
        if not isinstance(other, InfluenceGraph):
            return False
        return (self.genes == other.genes
                and self.process == other.process)



# -*- coding: utf8 -*-

from typing import List, Iterable, Tuple
import re

from .influence_graph import InfluenceGraph
from .discrete_model import DiscreteModel


def parse_smbionet_output_file(filename: str) -> Tuple[DiscreteModel]:
    with open(filename) as output:
        return parse_smbionet_output_string(output.read())


def parse_smbionet_output_string(string: str) -> Tuple[DiscreteModel]:
    graph = InfluenceGraph()
    match = _create_pattern().match(string)
    _parse_genes(match['genes'], graph)
    _parse_process(match['process'], graph)
    return tuple(_parse_models(match['models'], graph))


def _create_pattern() -> re.Pattern:
    read_genes = r'\s*VAR(?P<genes>(?:.|\s)*?)'
    read_process = r'REG(?P<process>(?:.|\s)*?)'
    read_parameters = r'(?:PARA(?P<parameters>(?:.|\s)*?))?'
    read_ctl = r'(?:CTL(?P<ctl>(?:.|\s)*?))?'
    read_models = r'(?P<models>#(?:.|\s)*?)'
    read_check = r'(?P<checks>#\s*SELECTED MODELS(?:.|\s)*?)?'
    return re.compile(
        '^'
        + read_genes
        + read_process
        + read_parameters
        + read_ctl
        + read_models
        + read_check
        + '$'
    )


def _parse_genes(string: str, graph: InfluenceGraph):
    pattern = re.compile(
        r'\s*(?P<gene>\w+)\s*=\s*(?P<start>\d+)\s+(?P<end>\d+)\s*;')
    for match in pattern.finditer(string):
        graph.add_gene(match['gene'], int(
            match['start']), int(match['end']))


def _parse_process(string: str, graph: InfluenceGraph):
    pattern = re.compile(
        r'(?P<name>\w+)\s+\[(?P<predicate>.*?)\]\s*=>\s*(?P<genes>(?:\w+\s*)+)\s*;')
    for match in pattern.finditer(string):
        graph.add_process(match['name'], _parse_predicate(match['predicate']),
                          *match['genes'].split())


def _parse_predicate(predicate: str):
    predicate = re.sub(r'(?<![<>])=', '==', predicate)
    predicate = re.sub(r'!', 'not', predicate)
    predicate = re.sub(r'\|', 'or', predicate)
    predicate = re.sub(r'&', 'and', predicate)
    return predicate


def _parse_models(string: str, graph: InfluenceGraph) -> Iterable[DiscreteModel]:
    pattern = re.compile(
        r'#\s+MODEL\s+(?P<id>\d+)\s+(?P<k_assigments>(?:(?!\s*#\s*MODEL).*\s*)*)')
    for match in pattern.finditer(string):
        yield _parse_k_assigments(match['k_assigments'], graph)


def _parse_k_assigments(string: str, graph: InfluenceGraph) -> DiscreteModel:
    model = DiscreteModel(graph)
    pattern = re.compile(
        r'#\s*K_(?P<gene>[^\s=]+)\s*=\s*(?P<states>(?:\d+\s+)*?\d+)\s*$', flags=re.MULTILINE)
    for match in pattern.finditer(string):
        gene, *context = match['gene'].split('+')
        model.add_transition(gene, tuple(context),
                             tuple(map(int, match['states'].split())))
    return model



from .parser import parse_smbionet_output_string, parse_smbionet_output_file
from .influence_graph import InfluenceGraph, Gene, Process
from .discrete_model import DiscreteModel, Transition


__all__ = [
    'Gene',
    'Process',
    'Transition',
    'DiscreteModel',
    'InfluenceGraph',
    'parse_smbionet_output_string',
    'parse_smbionet_output_file'
    ]
